/*  Xm headers  */
#include <Xm/DialogS.h>
/*
#include <Xm/RowColumn.h>
*/
#include "Grid.h"
#include <Xm/RowColumn.h>
#include <Xm/Frame.h>
#include <Xm/Label.h>
#include <Xm/PushBG.h>
#include <Xm/CascadeB.h>
#include <Xm/MessageB.h>
#include <Xm/MainW.h>
#include "Knob.h"

/* geometry */
#define HORIZONTAL_SPACING         (Dimension) 30
#define VERTICAL_SPACING           5
#define TOP_OFFSET                 (Dimension) 5
#define KNOB_HIGHLIGHT_THICKNESS  (Dimension) 2

#define RED     	1
#define GREEN   	2
#define BLUE    	3
#define ALL_COLORS	4

#define BUFFERSZ       1000

Widget  appshell;
Widget  frame; 
Widget  grid; 
Widget  redLabel; 
Widget  greenLabel; 
Widget  blueLabel; 
Widget  redKnob; 
Widget  greenKnob; 
Widget  blueKnob; 
Widget  redValueLabel; 
Widget  greenValueLabel; 
Widget  blueValueLabel; 

XColor  bg_color;   /* color structure for the background color */
XColor  fg_color;   /* color structure for the foreground color */
XColor  ts_color;   /* color structure for the top shadow color */
XColor  bs_color;   /* color structure for the bottom shadow color */ 
XColor  sc_color;   /* color structure for the select color (arm) color */ 

XmColorProc calcRGB;

Display *display;

Colormap colormap;

/* Forward Declarations */
static Widget CreateHelp(Widget);
static void HelpCB(Widget, XtPointer, XtPointer);
static void QuitCB(Widget, XtPointer, XtPointer);
static XtArgVal GetColor(char *);
static void CreateColorEditor(void);
static void ChangeRGB_CB(Widget, XtPointer, XtPointer);
static void SetKnobs(XColor *, int);
static void GenerateColors(void);
static void SetColors(Widget, char *);

Pixel GetPixel(Widget, char *);

char *tmpbuf = "#78a0d5";
/****************************************/
/* main                                 */
/****************************************/
void main(int argc, char ** argv)
{
   register int n;
   Arg args[4];
   Widget main_window, menu_bar, menu_pane, cascade;
   Widget button;
   Boolean status;
   unsigned long plane_mask;
   unsigned long *color_cells;
   XtAppContext app_context;

 /* Initialize the toolkit and open the display */
   n = 0;
   XtSetArg (args[n], XmNallowShellResize, True); n++;
   appshell = XtAppInitialize(&app_context, "Knobcolors", NULL, 0, &argc,
                              argv, NULL, args, n);

 /* Get the display */
   display = XtDisplay(appshell);

 /* Get the default colormap for this display */
   colormap = XDefaultColormap(display, XDefaultScreen(display));

 /* Allocate 4 color cells for use with this program: background, 
    foreground, top shadow, and bottom shadow.  The select color 
    won't be used so we don't need to allocate for it */
   color_cells = (unsigned long *)XtMalloc (4 * sizeof (unsigned long));
   status = XAllocColorCells (display, colormap,
                                     0, &plane_mask, 0, color_cells, 4);

   if(status == False)
   {
      printf(" couldn't allocate enough color cells\n");
      exit(1);
   }
   else
   {
 /* Assign the allocated color cells to the correct structure */
      n = 0;
      bg_color.pixel = color_cells[n++];
      bg_color.flags = DoRed | DoGreen | DoBlue;
      fg_color.pixel = color_cells[n++];
      fg_color.flags = DoRed | DoGreen | DoBlue;
      bs_color.pixel = color_cells[n++];
      bs_color.flags = DoRed | DoGreen | DoBlue;
      ts_color.pixel = color_cells[n];
      ts_color.flags = DoRed | DoGreen | DoBlue;
   }
 
 /* Get the RGB value (XColor) of the background */
   XParseColor(display, colormap, tmpbuf, &bg_color);
 
 /* Create main window. */
   main_window = XmCreateMainWindow (appshell, "main1", args, 0);
   XtManageChild (main_window);
 
 /* Create menu bar in main window. */
   menu_bar = XmCreateMenuBar (main_window, "menu_bar", args, 0); 
   XtManageChild (menu_bar);
 
 /* Create "Exit" pulldown menu. */
   menu_pane = XmCreatePulldownMenu (menu_bar, "menu_pane", args, 0);
 
   button = XmCreatePushButtonGadget (menu_pane, "Quit", args, 0);
   XtManageChild (button);
   XtAddCallback (button, XmNactivateCallback, QuitCB, NULL);
  
   n = 0;
   XtSetArg (args[n], XmNsubMenuId, menu_pane); n++;
   cascade = XmCreateCascadeButton (menu_bar, "Exit", args, n);
   XtManageChild (cascade);
 
 /* Create "Help" button. */
   cascade = XmCreateCascadeButton (menu_bar, "Help", args, 0);
   XtManageChild (cascade);
   XtAddCallback (cascade, XmNactivateCallback, HelpCB, NULL);
 
   n = 0;
   XtSetArg (args[n], XmNmenuHelpWidget, cascade); n++;
   XtSetValues (menu_bar, args, n);
 
 /*
  * Create a frame widget
  */
   n = 0;
   XtSetArg(args[n], XmNmarginWidth, 15); n++;
   XtSetArg(args[n], XmNmarginHeight, 15); n++;
   frame = XmCreateFrame (main_window, "frame", args, n);
   XtManageChild(frame);
 
 /*  Set main window areas  */
   XmMainWindowSetAreas (main_window, menu_bar, NULL, NULL, NULL, frame);
 
 /* Go generate the foreground, top shadow, and bottom shadow colors
    based on the background color which was set by XParseColor above */
   GenerateColors();
 
 /* Create the sliders with labels using the generated colors */
   CreateColorEditor();
 
   XtRealizeWidget (appshell);

   XtAppMainLoop (app_context);
}


/*******************************************
 *  HelpCB      - callback for help button *
 ********************************************/
/* ARGSUSED */
static void
HelpCB(Widget w, XtPointer client_data, XtPointer call_data) 
{
   Widget   message_box;  /*  message box             */

 /*  Create help window. */
   message_box = CreateHelp (w);

 /*  Display help window. */
   XtManageChild (message_box);
}


/*******************************************
 *  CreateHelp    - create help window
 *******************************************/
static Widget
CreateHelp(Widget parent) 
{
   Widget        button;
   Widget        message_box;       /*  Message Dialog  */
   Arg           args[4];           /*  arg list        */
   register int  n;                 /*  arg count       */

   static char   message[BUFFERSZ];  /*  help text      */
   XmString      title_string = NULL;
   XmString      message_string = NULL;
   XmString      button_string = NULL;


 /*  Generate message to display. */
 
   sprintf (message, "\
This program uses three knob widgets to balance colors. The three knob\n\
are used to adjust the primary colors: red, blue, and green.  As the\n\
knob are adjusted, the background color changes accordingly.  To\n\
terminate the program, press the 'Exit' button and then the 'Quit'\n\
button.");


 /* Create the compound strings */
   message_string = XmStringCreateLocalized (message);
   button_string = XmStringCreateLocalized ("Close");
   title_string = XmStringCreateLocalized ("knobcolors help");
 
 /*  Create message box dialog. */
   n = 0;
   XtSetArg (args[n], XmNdialogTitle, title_string);  n++;
   XtSetArg (args[n], XmNokLabelString, button_string);  n++;
   XtSetArg (args[n], XmNmessageString, message_string);  n++;
   message_box = XmCreateMessageDialog (parent, "helpbox", args, n);

   button = XmMessageBoxGetChild (message_box, XmDIALOG_CANCEL_BUTTON);
   XtUnmanageChild (button);
   button = XmMessageBoxGetChild (message_box, XmDIALOG_HELP_BUTTON);
   XtUnmanageChild (button);

 /*  Free strings and return message box. */
   if (title_string) XmStringFree (title_string);
   if (message_string) XmStringFree (message_string);
   if (button_string) XmStringFree (button_string);
   return (message_box);
}


/*******************************************
 *  QuitCB      - callback for quit button
 *******************************************/
/* ARGSUSED */
static void
QuitCB (Widget w, XtPointer client_data, XtPointer call_data) 
{
 /*  Terminate the application. */
   exit (0);
}

/******************************************************************
 *  GetColor      - function for converting a string into a color
 ******************************************************************/
static XtArgVal
GetColor(char * colorstr)
{
   XrmValue from, to;

   from.size = strlen(colorstr) +1;
   if (from.size < sizeof(String)) from.size = sizeof(String);
   from.addr = colorstr;
   to.addr = NULL;
   XtConvert(appshell, XmRString, &from, XmRPixel, &to);

   return ((XtArgVal) *((XtArgVal *) to.addr));
}


/**********************************************************
 * CreateColorEditor  -  Create the grid and knob widgets
 **********************************************************/
static void 
CreateColorEditor(void)
{
   register int     n;
   Arg              args[15];
   XmString         string;               /* temp Xm string */
   XColor *rgb = &bg_color;
   char* color_string = XtMalloc(sizeof(char)*5);
   int new_color;


 /* Create grid widget */
   n=0;
   XtSetArg(args[n], XmNbackground, bg_color.pixel);  n++;
   XtSetArg(args[n], XmNforeground, fg_color.pixel);  n++;
   XtSetArg(args[n], XmNtopShadowColor, ts_color.pixel);  n++;
   XtSetArg(args[n], XmNbottomShadowColor, bs_color.pixel);  n++;
   XtSetArg(args[n], XsmNhorizontalAlignment, XsmALIGN_CENTER);  n++;
   XtSetArg(args[n], XsmNshowGridLines, True);  n++;
   XtSetArg(args[n], XsmNrows, 3);  n++;
   XtSetArg(args[n], XsmNcolumns, 3);  n++;
   grid = XsmCreateGrid(frame, "grid", args, n);
                      
   XtManageChild (grid);
 
 /*
  * Create the three groups widgets, one each for red, green, and blue
  */

  /* Create red label, knob, values & label */
   new_color = rgb->red >> 8;

   string =  XmStringCreateLocalized("R");
   n = 0;
   XtSetArg(args[n], XmNlabelString, string); n++;
   XtSetArg(args[n], XsmNrowPosition, 0); n++; 
   XtSetArg(args[n], XsmNcolumnPosition, 0); n++; 
   redLabel = XmCreateLabel(grid, "redLabel", args, n);

   XtManageChild(redLabel);
   XmStringFree(string);
 
   n = 0;
   XtSetArg(args[n], XmNbackground, bg_color.pixel);  n++;
   XtSetArg(args[n], XmNforeground, GetColor("maroon")); n++; 
   XtSetArg(args[n], XsmNmaxValue, 0xff);  n++;
   XtSetArg(args[n], XsmNminValue, 0x0);  n++;
   XtSetArg(args[n], XsmNknobColor, GetColor("red")); n++; 
   XtSetArg(args[n], XsmNindicatorColor, GetColor("Blue")); n++; 
   XtSetArg(args[n], XsmNvalue, new_color); n++; 
   XtSetArg(args[n], XsmNrowPosition, 1); n++; 
   XtSetArg(args[n], XsmNcolumnPosition, 0); n++; 
   XtSetArg(args[n], XsmNturnDelay, 70); n++; 
   XtSetArg(args[n], XsmNmarginWidth, 1); n++; 
   XtSetArg(args[n], XsmNmarginHeight, 1); n++; 
   redKnob = XsmCreateKnob(grid, "redKnob", args, n);

   XtManageChild(redKnob);
   XtAddCallback(redKnob, XsmNvalueChangedCallback, ChangeRGB_CB,
                                                      (XtPointer) RED);
   n = 0;
   sprintf(color_string, "%d", new_color);

   string =  XmStringCreateLocalized(color_string);
   XtSetArg(args[0], XmNlabelString, string); n++;
   XtSetArg(args[n], XsmNrowPosition, 2); n++; 
   XtSetArg(args[n], XsmNcolumnPosition, 0); n++; 
   redValueLabel = XmCreateLabel(grid, "redValueLabel", args, n);

   XtManageChild(redValueLabel);
   XmStringFree(string);

  /* Create green label, knob, values & label */
   new_color = rgb->green >> 8;

   string =  XmStringCreateLocalized("G");
   n = 0;
   XtSetArg(args[n], XmNlabelString, string); n++;
   XtSetArg(args[n], XsmNrowPosition, 0); n++; 
   XtSetArg(args[n], XsmNcolumnPosition, 1); n++; 
   greenLabel = XmCreateLabel(grid, "greenLabel", args, n);

   XtManageChild(greenLabel);
   XmStringFree(string);

   n = 0;
   XtSetArg(args[n], XmNbackground, bg_color.pixel);  n++;
   XtSetArg(args[n], XmNforeground, GetColor("ForestGreen")); n++; 
   XtSetArg(args[n], XsmNmaxValue, 0xff);  n++;
   XtSetArg(args[n], XsmNminValue, 0x0);  n++;
   XtSetArg(args[n], XsmNknobColor, GetColor("Green")); n++; 
   XtSetArg(args[n], XsmNindicatorColor, GetColor("Red")); n++; 
   XtSetArg(args[n], XsmNvalue, new_color); n++; 
   XtSetArg(args[n], XsmNrowPosition, 1); n++; 
   XtSetArg(args[n], XsmNcolumnPosition, 1); n++; 
   XtSetArg(args[n], XsmNturnDelay, 70); n++; 
   XtSetArg(args[n], XsmNmarginWidth, 1); n++; 
   XtSetArg(args[n], XsmNmarginHeight, 1); n++; 
   greenKnob = XsmCreateKnob(grid, "greenKnob", args, n);

   XtManageChild(greenKnob);
   XtAddCallback(greenKnob, XsmNvalueChangedCallback, ChangeRGB_CB,
                                                      (XtPointer) GREEN );
   n = 0;
   sprintf(color_string, "%d", new_color);
   string =  XmStringCreateLocalized(color_string);
   XtSetArg(args[0], XmNlabelString, string); n++;
   XtSetArg(args[n], XsmNrowPosition, 2); n++; 
   XtSetArg(args[n], XsmNcolumnPosition, 1); n++; 
   greenValueLabel = XmCreateLabel(grid, "greenValueLabel", args, n);

   XtManageChild(greenValueLabel);
   XmStringFree(string);

  /* Create blue label, knob, values & label */
   new_color = rgb->blue >> 8;
 
   string =  XmStringCreateLocalized("B");
   n = 0;
   XtSetArg(args[n], XmNlabelString, string); n++;
   XtSetArg(args[n], XsmNrowPosition, 0); n++; 
   XtSetArg(args[n], XsmNcolumnPosition, 2); n++; 
   blueLabel = XmCreateLabel(grid, "blueLabel", args, n);

   XtManageChild(blueLabel);
   XmStringFree(string);

   n = 0;
   XtSetArg(args[n], XmNbackground, bg_color.pixel);  n++;
   XtSetArg(args[n], XmNforeground, GetColor("navy")); n++; 
   XtSetArg(args[n], XsmNmaxValue, 0xff);  n++;
   XtSetArg(args[n], XsmNminValue, 0x0);  n++;
   XtSetArg(args[n], XsmNknobColor, GetColor("Blue")); n++; 
   XtSetArg(args[n], XsmNindicatorColor, GetColor("red")); n++; 
   XtSetArg(args[n], XsmNvalue, new_color); n++; 
   XtSetArg(args[n], XsmNrowPosition, 1); n++; 
   XtSetArg(args[n], XsmNcolumnPosition, 2); n++; 
   XtSetArg(args[n], XsmNturnDelay, 70); n++; 
   XtSetArg(args[n], XsmNmarginWidth, 1); n++; 
   XtSetArg(args[n], XsmNmarginHeight, 1); n++; 
   blueKnob = XsmCreateKnob(grid, "blueKnob", args, n);

   XtManageChild(blueKnob);
   XtAddCallback(blueKnob, XsmNvalueChangedCallback, ChangeRGB_CB,
                                                      (XtPointer) BLUE );
   n = 0;
   sprintf(color_string, "%d", new_color);
   string =  XmStringCreateLocalized(color_string);
   XtSetArg(args[0], XmNlabelString, string); n++;
   XtSetArg(args[n], XsmNrowPosition, 2); n++; 
   XtSetArg(args[n], XsmNcolumnPosition, 2); n++; 
   blueValueLabel = XmCreateLabel(grid, "blueValueLabel", args, n);

   XtManageChild(blueValueLabel);
   XmStringFree(string);

   XtFree(color_string);

}

/************************************************************************
 *   GenerateColors()
 *           Generates new RGB values for foreground, top shadow, 
 *           bottom shadow, and select color based on the background.
 *           Select color is not used in our program but the Motif color
 *           generation routine expects it and calculates it.
 *           The Colors generated are stored in the red, green, and blue
 *           fields of the XColor structure passed in.
 *           The generated colors are then used to update the color cells
 *           of the ColorSet being edited.
 ************************************************************************/
static void 
GenerateColors(void)
{
   XColor colors[4];
   int j=0;
 
   if (calcRGB == NULL) 
       calcRGB = XmGetColorCalculation();
 
 /* Given the background color, calculate the foreground, select color,
     top shadow, and bottom shadow colors */
   (*calcRGB)(&bg_color, &fg_color, &sc_color, &ts_color, &bs_color);
   
 /* Put those calculated colors in an array */
   colors[j++] =  bg_color;
   colors[j++] =  fg_color; 
   colors[j++] =  ts_color;
   colors[j++] =  bs_color;
 
 /* Go change the colors dynamically */
   XStoreColors(display, colormap, colors, j);
}

/************************************************************************
 *   ChangeRGB_CB()
 *           Called when one of the RGB knobs is moved
 ************************************************************************/
/* ARGSUSED */
static void 
ChangeRGB_CB( Widget w, XtPointer client_data, XtPointer call_data )
{
   int reason_code;
   int value;
   int color;

   reason_code = ((XmAnyCallbackStruct *)call_data)->reason;
   if ( reason_code == XsmCR_VALUE_CHANGED)
   {
      color = (int) client_data;
      value = ((XsmKnobCallbackStruct *)call_data)->value;

      /*
       * Shift value -- to make up for knob max of only 0xff
       */
      value <<= 8;
      switch (color)
      {
        case RED:
           bg_color.red = value;
           break;
        case GREEN:
           bg_color.green = value;
           break;
        case BLUE:
           bg_color.blue = value;
           break;
        default:
           return;
      }

      SetKnobs(&bg_color, color);
      GenerateColors();
   }
}

/************************************************************************
 * SetColors()
 *          change the label string in the value label widget
 ************************************************************************/
static void
SetColors(Widget valueLabel, char * color_string)
{
   Arg      args[1];
   XmString string;

   string =  XmStringCreateLocalized(color_string);
   XtSetArg(args[0], XmNlabelString, string);
   XtSetValues(valueLabel, args, 1);
   XmStringFree(string);
}

/************************************************************************
 *   SetKnobs()
 *           passed an XColor, updates all knobs.
 ************************************************************************/
static void 
SetKnobs(XColor * rgb, int color)
{
   char* color_string = XtMalloc(sizeof(char)*5);
   int new_color;

   switch (color)
   {
        case RED:
   		new_color = rgb->red >> 8;
   		sprintf(color_string, "%d", new_color);
		SetColors(redValueLabel, color_string);
		break;

        case GREEN:
   		new_color = rgb->green >> 8;
      		sprintf(color_string, "%d", new_color);
		SetColors(greenValueLabel, color_string);
		break;

        case BLUE:
   		new_color = rgb->blue >> 8;
      		sprintf(color_string, "%d", new_color);
		SetColors(blueValueLabel, color_string);
		break;

	default:
		break;
   }

   XtFree(color_string);
}


/************************************************************************
 *   GetPixel()
 *         passed in a color string, calls XtConvert which is an X 
 *         toolkit converter which converts a string into a pixel value.
 ************************************************************************/
Pixel 
GetPixel( Widget widget, char * color_string )
{
   XrmValue from, to;

 /* Determine the size of the string passed in */
   from.size = strlen(color_string) + 1;
   if (from.size < sizeof(String))
       from.size = sizeof(String);
   from.addr = color_string;

 /* Go convert the string into a pixel (i.e. color cell) */
   XtConvert(widget, XmRString, &from, XmRPixel, &to);

 /* Return that pixel */
   return ((Pixel) *((Pixel *) to.addr));
}

